@@ -18,11 +18,18 @@ class DBException(Exception): #{{{1
 
 class DBExceptionPrepare(DBException): #{{{1
     def __init__(self, query):
-        DBException.__init("Error preparing query", query.lastError().text())
+        DBException.__init__(self, "Error preparing query", query.lastError().text())
 
 class DBExceptionExecute(DBException): #{{{1
     def __init__(self, query):
-        DBException.__init("Error executing query", query.lastError().text())
+        DBException.__init__(self, "Error executing query", query.lastError().text())
+
+class DBExceptionData(Exception): #{{{1
+    def __init__(self, data):
+        self.data = data
+
+    def __str__(self):
+        return repr("Unrecognised data: %s" % self.data)
 
 class LTDatabase(): #{{{1
     # Class variables {{{2
@@ -34,13 +41,15 @@ class LTDatabase(): #{{{1
 
     # Define the SQL queries #{{{2
     SQL_CREATE_DISPLAY_OPTIONS = 'create table display_options ( column text, display_name text, display_order integer, sort_order integer, display_width integer)'
+    SQL_CREATE_DISPLAY_FILTERS = 'create table display_filters ( filter_name text, filter_column text, filter_method text, filter_string text)'
 
     SQL_DELETE_BOOKS = 'delete from books'
 
     SQL_INSERT_DISPLAY_OPTION = 'insert into display_options values(:col, :name, :disp, :sort, 0)'
 
-    SQL_SELECT_DISPLAY_ORDER = 'select column, display_name, display_width from display_options where display_order > 0 order by display_order'
+    SQL_SELECT_DISPLAY_COLUMNS = 'select column from display_options where display_order > 0 order by display_order'
     SQL_SELECT_DISPLAY_SORT = 'select column from display_options where sort_order > 0 order by sort_order'
+    SQL_SELECT_DISPLAY_WIDTHS = 'select display_width from display_options where display_order > 0 order by display_order'
 
     SQL_SELECT_FIELDS_DISPLAY = 'select display_name from display_options where display_order > 0 order by display_order'
     SQL_SELECT_FIELDS_NONDISPLAY = "select display_name from display_options where display_order = 0 and column != 'id'"
@@ -50,6 +59,9 @@ class LTDatabase(): #{{{1
     SQL_SELECT_FIELD_NAMES = 'select column, display_name from display_options'
     SQL_SELECT_BOOK_INFO = 'select * from books where id = :id'
 
+    SQL_SELECT_FILTER = 'select display_name, filter_method, filter_string from display_filters, display_options where column = filter_column and filter_name = :name'
+    SQL_SELECT_CURRENT_FILTER = "select filter_column, filter_method, filter_string from display_filters where filter_name = 'current'"
+
     SQL_UPDATE_DISPLAY_ORDER_1 = 'update display_options set display_order = display_order - 1 where display_order > %d and display_order <= %d'
     SQL_UPDATE_DISPLAY_ORDER_2 = 'update display_options set display_order = display_order + 1 where display_order >= %d and display_order < %d'
     SQL_UPDATE_DISPLAY_ORDER_3 = 'update display_options set display_order = :pos where column = :col'
@@ -60,6 +72,13 @@ class LTDatabase(): #{{{1
     SQL_RESET_DISPLAY_ORDER = 'update display_options set display_order = 0'
     SQL_RESET_SORT_ORDER = 'update display_options set sort_order = 0'
 
+    SQL_FILTER_DELETE = 'delete from display_filters where filter_name = :name'
+    SQL_FILTER_RENAME = 'update display_filters set filter_name = :new where filter_name = :old'
+    SQL_FILTER_COPY   = 'insert into display_filters (filter_name, filter_column, filter_method, filter_string) select :new, filter_column, filter_method, filter_string from display_filters where filter_name = :old'
+    SQL_FILTER_ADD_RULE = 'insert into display_filters (filter_name, filter_column, filter_method, filter_string) values (:name, :column, :method, :string)'
+    SQL_FILTER_DEL_RULE = 'delete from display_filters where filter_name = :name and filter_column = :column and filter_method = :method and filter_string = :string'
+    SQL_FILTER_LIST = "select distinct filter_name from display_filters where filter_name not in ('current', 'working')"
+
     # Define the fields to be used (mapped from LibraryThing export) #{{{2
     bookFields = [
         [ "id", "integer", "ID" ],
@@ -104,11 +123,6 @@ class LTDatabase(): #{{{1
         if not configDir.exists():
             raise IOError(errno.ENOTDIR, "Not a directory", str(configDir.path()).encode())
 
-        # Check whether we're creating a new database
-        new_db = False
-        if not configDir.exists(self._dbFile):
-            new_db = True
-
         # Open/Create the SQLite database file
         self._conn = QSqlDatabase.addDatabase('QSQLITE')
         dbFile = QDir.toNativeSeparators(configDir.path() + "/" + self._dbFile)
@@ -117,11 +131,19 @@ class LTDatabase(): #{{{1
         if not self._conn.open():
             raise IOError(errno.EIO, "Failed to open database", str(dbFile).encode())
 
-        # If this is a new database, initialise it
-        if new_db:
-            self._initialise()
+        # Get a list of tables and create any new ones needed
+        tables = self._conn.tables()
 
-    def _initialise(self): #{{{2
+        if not tables.contains(QString('display_options')):
+            self._createTableDisplayOptions()
+
+        if not tables.contains(QString('books')):
+            self._createTableBooks()
+
+        if not tables.contains(QString('display_filters')):
+            self._createTableDisplayFilters()
+
+    def _createTableDisplayOptions(self): #{{{2
         # Create the display options table
         q = QSqlQuery(self.SQL_CREATE_DISPLAY_OPTIONS, self._conn)
         if not q.isActive():
@@ -152,6 +174,7 @@ class LTDatabase(): #{{{1
 
         self._conn.commit()
 
+    def _createTableBooks(self): #{{{2
         # Create the books table
         createSql = 'create table books('
 
@@ -166,6 +189,11 @@ class LTDatabase(): #{{{1
         if not q.isActive():
             raise DBExceptionExecute(q)
 
+    def _createTableDisplayFilters(self): #{{{2
+        q = QSqlQuery(self.SQL_CREATE_DISPLAY_FILTERS, self._conn)
+        if not q.isActive():
+            raise DBExceptionExecute(q)
+
     def loadFile(self, filename, progress = None): #{{{2
         file = QFile(filename)
 
@@ -279,23 +307,64 @@ class LTDatabase(): #{{{1
         self._conn.commit()
 
     def getQuery(self, model): #{{{2
+        columnSql = self._getQueryColumns()
+        orderSql  = self._getQueryOrder()
+        (filterSql, filterStrings) = self._getQueryFilter()
+
+        querySql  = columnSql
+
+        if len(filterSql) > 0:
+            querySql += ' ' + filterSql
+
+        querySql += ' ' + orderSql
+
+        if len(filterSql) > 0:
+            q = QSqlQuery(self._conn)
+            
+            if not q.prepare(querySql):
+                raise DBExceptionPrepare(q)
+                
+            for i, val in enumerate(filterStrings):
+                bindVariable = ':%d' % i
+                q.bindValue(bindVariable, QVariant(val))
+
+            if not q.exec_():
+                raise DBExceptionExecute(q)
+
+            querySql = q
+
+        # Create the query model
+        model.setQuery(querySql)
+
+        # Ensure all data is retrieved
+        while model.canFetchMore():
+            model.fetchMore()
+
+        # Set the column headers
+        colNames = self.getDisplayColumns()
+
+        for i, colName in enumerate(colNames):
+            model.setHeaderData(i + 1, Qt.Horizontal, QVariant(colName))
+
+    def _getQueryColumns(self): #{{{2
         querySql = QString('select id, ')
-        colNames = []
-        colWidths = []
 
         # Lookup the columns to display, along with the names and widths
-        q = QSqlQuery(self.SQL_SELECT_DISPLAY_ORDER, self._conn)
+        q = QSqlQuery(self.SQL_SELECT_DISPLAY_COLUMNS, self._conn)
         if not q.isActive():
             raise DBExceptionExecute(q)
 
         while q.next():
             res = q.result().data(0).toString()
             querySql += res + ', '
-            colNames.append(q.result().data(1))
-            colWidths.append(q.result().data(2).toInt()[0])
 
         querySql.chop(2)
-        querySql += ' from books order by '
+        querySql += ' from books'
+
+        return querySql
+
+    def _getQueryOrder(self): #{{{2
+        querySql = QString('order by ')
 
         # Look up the ordering for the query results
         q = QSqlQuery(self.SQL_SELECT_DISPLAY_SORT, self._conn)
@@ -313,18 +382,44 @@ class LTDatabase(): #{{{1
         else:
             querySql += 'id'
 
-        # Create the query model
-        model.setQuery(querySql)
+        return querySql
 
-        # Ensure all data is retrieved
-        while model.canFetchMore():
-            model.fetchMore()
+    def _getQueryFilter(self): #{{{2
+        querySql = QString('where ')
+        filterStrings = []
 
-        # Set the column headers
-        for i, colName in enumerate(colNames):
-            model.setHeaderData(i + 1, Qt.Horizontal, colName)
+        # Look up the current query filter
+        q = QSqlQuery(self.SQL_SELECT_CURRENT_FILTER, self._conn)
+        if not q.isActive():
+            raise DBExceptionExecute(q)
 
-        return colWidths
+        filterSeen = False
+        i = 0
+        while q.next():
+            filterSeen = True
+            res = q.result()
+            col = res.data(0).toString()
+            method = res.data(1).toString()
+            f_string = res.data(2).toString()
+
+            if method == '=':
+                querySql += 'upper(%s) = upper(:%d) and ' % (col, i)
+                filterStrings.append(f_string)
+            elif method == 'contains':
+                querySql += 'upper(%s) like upper(:%d) and ' % (col, i)
+                filterStrings.append('%' + f_string + '%')
+            elif method == 'begins with':
+                querySql += 'upper(%s) like upper(:%d) and ' % (col, i)
+                filterStrings.append(f_string + '%')
+            else:
+                raise DBExceptionData(method)
+
+        if filterSeen:
+            querySql.chop(5)
+        else:
+            querySql = ''
+
+        return (querySql, filterStrings)
 
     def changeColumnOrder(self, col, oldPos, newPos): #{{{2
         self._conn.transaction()
@@ -367,6 +462,14 @@ class LTDatabase(): #{{{1
     def getNonDisplayColumns(self): #{{{2
         return self.getColumns(self.SQL_SELECT_FIELDS_NONDISPLAY)
 
+    def getColumnWidths(self): #{{{2
+        colWidths = []
+
+        for val in self.getColumns(self.SQL_SELECT_DISPLAY_WIDTHS):
+            colWidths.append(val.toInt()[0])
+
+        return colWidths
+
     def getDisplayColumns(self): #{{{2
         return self.getColumns(self.SQL_SELECT_FIELDS_DISPLAY)
 
@@ -471,6 +574,77 @@ class LTDatabase(): #{{{1
 
         return bookData
 
+    def getFilter(self, model, name): #{{{2
+        q = QSqlQuery(self._conn)
+        if not q.prepare(self.SQL_SELECT_FILTER):
+            raise DBExceptionPrepare(q)
+
+        q.bindValue(':name', QVariant(name))
+        if not q.exec_():
+            raise DBExceptionExecute(q)
+
+        model.setQuery(q)
+
+    def filterAddRule(self, name, field, method, f_string): #{{{2
+        q = QSqlQuery(self._conn)
+        if not q.prepare(self.SQL_FILTER_ADD_RULE):
+            raise DBExceptionPrepare(q)
+
+        q.bindValue(':name', QVariant(name))
+        q.bindValue(':column', QVariant(field))
+        q.bindValue(':method', QVariant(method))
+        q.bindValue(':string', QVariant(f_string))
+        if not q.exec_():
+            raise DBExceptionExecute(q)
+
+    def filterRemoveRule(self, name, field, method, f_string): #{{{2
+        q = QSqlQuery(self._conn)
+        if not q.prepare(self.SQL_FILTER_DEL_RULE):
+            raise DBExceptionPrepare(q)
+
+        q.bindValue(':name', QVariant(name))
+        q.bindValue(':column', QVariant(field))
+        q.bindValue(':method', QVariant(method))
+        q.bindValue(':string', QVariant(f_string))
+        if not q.exec_():
+            raise DBExceptionExecute(q)
+
+    def filterRename(self, oldName, newName): #{{{2
+        self.filterDelete(newName)
+
+        q = QSqlQuery(self._conn)
+        if not q.prepare(self.SQL_FILTER_RENAME):
+            raise DBExceptionPrepare(q)
+
+        q.bindValue(':old', QVariant(oldName))
+        q.bindValue(':new', QVariant(newName))
+        if not q.exec_():
+            raise DBExceptionExecute(q)
+
+    def filterDelete(self, name): #{{{2
+        q = QSqlQuery(self._conn)
+        if not q.prepare(self.SQL_FILTER_DELETE):
+            raise DBExceptionPrepare(q)
+
+        q.bindValue(':name', QVariant(name))
+        if not q.exec_():
+            raise DBExceptionExecute(q)
+
+    def filterCopy(self, oldName, newName): #{{{2
+        self.filterDelete(newName)
+
+        q = QSqlQuery(self._conn)
+        if not q.prepare(self.SQL_FILTER_COPY):
+            raise DBExceptionPrepare(q)
+
+        q.bindValue(':old', QVariant(oldName))
+        q.bindValue(':new', QVariant(newName))
+        if not q.exec_():
+            raise DBExceptionExecute(q)
+
+    def filterList(self): #{{{2
+        return self.getColumns(self.SQL_FILTER_LIST)
+
 # Handle running as a script {{{1
 if __name__ == '__main__':
     app = QCoreApplication(sys.argv)
