@@ -2,18 +2,18 @@
 
 from PySide.QtCore    import Qt, QByteArray, QIODevice, QModelIndex, QString, \
                              QTextStream, QUrl, QVariant, SIGNAL
-from PySide.QtGui     import QAbstractItemView, QDialog, QDialogButtonBox, \
-                             QFileDialog, QGridLayout, QGroupBox, \
-                             QHBoxLayout, QInputDialog, QLabel, QLineEdit, \
-                             QListView, QMenuBar, QMessageBox, \
-                             QProgressDialog, QStringListModel, QTableView, \
-                             QToolButton, QVBoxLayout, QWidget
+from PySide.QtGui     import QAbstractItemView, QAction, QComboBox, QDialog, \
+                             QDialogButtonBox, QFileDialog, QGridLayout, \
+                             QGroupBox, QHBoxLayout, QInputDialog, QLabel, \
+                             QLineEdit, QListView, QMenuBar, QMessageBox, \
+                             QProgressDialog, QPushButton, QStringListModel, \
+                             QTableView, QToolButton, QVBoxLayout, QWidget
 from PySide.QtSql     import QSqlQueryModel
 from PySide.QtNetwork import QNetworkAccessManager, QNetworkCookieJar, \
                              QNetworkReply, QNetworkRequest
 from LTdatabase       import LTDatabase
 
-version = '0.0.2'
+version = '0.0.3'
 exportUrl = QUrl('http://www.librarything.com/export-tab')
 loginUrl  = QUrl('http://www.librarything.com/signup.php')
 
@@ -25,6 +25,149 @@ class LTException(Exception): #{{{1
     def __str__(self):
         return repr('%s: %s' % (self._source, self._error))
 
+class LTFilterDefine(QDialog): #{{{1
+    def __init__(self, parent, database): #{{{2
+        QDialog.__init__(self, parent)
+
+        self.setWindowTitle('Filters')
+        self._database = database
+
+        # Set up the list of filters {{{3
+        self._filters  = QTableView()
+        self._filters.setModel(QSqlQueryModel(self))
+        self._filters.setCornerButtonEnabled(False)
+        self._filters.verticalHeader().setVisible(False)
+        self._filters.verticalHeader().setDefaultSectionSize(40)
+        self._filters.horizontalHeader().setVisible(False)
+        self._filters.setSelectionMode(QAbstractItemView.NoSelection)
+
+        action = QAction('Remove Rule', self._filters)
+        action.connect(SIGNAL('triggered(bool)'), self.removeRule)
+        self._filters.addAction(action)
+        self._filters.setContextMenuPolicy(Qt.ActionsContextMenu)
+
+        # Set up the list of columns for filtering on {{{3
+        fieldNames   = database.getFieldNames()
+        self._revFieldNames = dict([(v, k) for (k, v) in fieldNames.iteritems()])
+
+        displayNames = self._revFieldNames.keys()
+        displayNames.sort()
+
+        self._filterFields = QComboBox()
+        for val in displayNames:
+            if self._revFieldNames[val] == 'id':
+                continue
+
+            self._filterFields.addItem(val, QVariant(self._revFieldNames[val]))
+
+        # Set up the list of filtering methods {{{3
+        self._filterMethods = QComboBox()
+        self._filterMethods.addItem(QString('='))
+        self._filterMethods.addItem(QString('contains'))
+        self._filterMethods.addItem(QString('begins with'))
+
+        # Set up the filter text field {{{3
+        self._filterString = QLineEdit()
+
+        # Set up the buttons {{{3
+        btnLoad  = QPushButton('Load Filter')
+        btnSave  = QPushButton('Save Filter')
+        btnClear = QPushButton('Clear Rules')
+        btnAdd   = QPushButton('Add Rule')
+
+        btnAdd.setDefault(True)
+
+        filterBtns = QDialogButtonBox(QDialogButtonBox.Ok|QDialogButtonBox.Cancel, Qt.Vertical)
+        filterBtns.addButton(btnLoad, QDialogButtonBox.ActionRole)
+        filterBtns.addButton(btnSave, QDialogButtonBox.ActionRole)
+        filterBtns.addButton(btnClear, QDialogButtonBox.ResetRole)
+
+        filterBtns.connect(SIGNAL('accepted()'), self.applyFilter)
+        filterBtns.connect(SIGNAL('rejected()'), self.cancelFilter)
+        btnClear.connect(SIGNAL('clicked(bool)'), self.clearFilter)
+        btnAdd.connect(SIGNAL('clicked(bool)'), self.addRule)
+        btnLoad.connect(SIGNAL('clicked(bool)'), self.loadFilter)
+        btnSave.connect(SIGNAL('clicked(bool)'), self.saveFilter)
+
+        # Layout all the elements {{{3
+        hbox1 = QHBoxLayout()
+        hbox1.addWidget(self._filterFields)
+        hbox1.addWidget(self._filterMethods)
+        hbox1.addWidget(self._filterString)
+        hbox1.addWidget(btnAdd)
+
+        hbox2 = QHBoxLayout()
+        hbox2.addWidget(self._filters)
+        hbox2.addWidget(filterBtns)
+
+        vbox = QVBoxLayout()
+        vbox.addLayout(hbox2)
+        vbox.addLayout(hbox1)
+
+        self.setLayout(vbox)
+
+        # Load the data {{{3
+        database.filterCopy('current', 'working')
+        database.getFilter(self._filters.model(), 'working')
+
+        self._filters.setColumnWidth(0, 200)
+        self._filters.setColumnWidth(1, 100)
+        self._filters.horizontalHeader().setStretchLastSection(True)
+
+    def addRule(self, checked): #{{{2
+        filterField = self._filterFields.itemData(self._filterFields.currentIndex())
+        filterMethod = self._filterMethods.currentText()
+        filterString = self._filterString.text()
+
+        if filterString.isEmpty():
+            return
+
+        self._database.filterAddRule('working', filterField, filterMethod, filterString)
+        self._database.getFilter(self._filters.model(), 'working')
+
+    def removeRule(self, bool): #{{{2
+        ruleRow = self._filters.currentIndex().row()
+
+        if ruleRow == -1:
+            return
+
+        model = self._filters.model()
+
+        filterField = self._revFieldNames[model.index(ruleRow, 0).data().toString()]
+        filterMethod = model.index(ruleRow, 1).data().toString()
+        filterString = model.index(ruleRow, 2).data().toString()
+        
+        self._database.filterRemoveRule('working', filterField, filterMethod, filterString)
+        self._database.getFilter(model, 'working')
+
+    def applyFilter(self): #{{{2
+        self._database.filterRename('working', 'current')
+        self.accept()
+
+    def cancelFilter(self): #{{{2
+        self._database.filterDelete('working')
+        self.reject()
+
+    def clearFilter(self, checked): #{{{2
+        self._database.filterDelete('working')
+        self._database.getFilter(self._filters.model(), 'working')
+
+    def loadFilter(self, checked): #{{{2
+        name = QInputDialog.getItem(self, 'Load filter', 'Select filter to load', self._database.filterList())[0]
+
+        if not name.isEmpty():
+            self._database.filterCopy(name, 'working')
+            self._database.getFilter(self._filters.model(), 'working')
+
+    def saveFilter(self, checked): #{{{2
+        if self._filters.model().rowCount() == 0:
+            return
+
+        name = QInputDialog.getText(self, 'Save filter', 'Enter name for filter')[0]
+
+        if not name.isEmpty():
+            self._database.filterCopy('working', name)
+
 class LTColumnSelectGeneric(QDialog): #{{{1
     def __init__(self, parent, database): #{{{2
         QDialog.__init__(self, parent)
@@ -89,7 +232,7 @@ class LTColumnSelectGeneric(QDialog): #{{{1
         # Add the Save/Cancel buttons #{{{3
         exitButtons = QDialogButtonBox(QDialogButtonBox.Save|QDialogButtonBox.Cancel,Qt.Horizontal)
         exitButtons.connect(SIGNAL('accepted()'), self.saveSettings)
-        exitButtons.connect(SIGNAL('rejected()'), self.reject)
+        exitButtons.connect(SIGNAL('rejected()'), self.cancelSelection)
 
         # Add all the widgets to the window {{{3
         selLyt.addWidget(exitButtons)
@@ -166,6 +309,9 @@ class LTColumnSelectGeneric(QDialog): #{{{1
 
         dstList.setCurrentIndex(newItem)
 
+    def cancelSelection(self): #{{{2
+        self.reject()
+
 class LTColumnSelectDisplay(LTColumnSelectGeneric): #{{{1
     def __init__(self, parent, database): #{{{2
         LTColumnSelectGeneric.__init__(self, parent, database)
@@ -291,6 +437,7 @@ class LTMainWindow(QWidget): #{{{1
         # Set up the menu
         menu.addAction('Display Fields').connect(SIGNAL('triggered()'), self.displayFields)
         menu.addAction('Sort Order').connect(SIGNAL('triggered()'), self.sortOrder)
+        menu.addAction('Filters').connect(SIGNAL('triggered()'), self.displayFilters)
         menu.addAction('Load Export File').connect(SIGNAL('triggered()'), self.loadFile)
         menu.addAction('Download Export File').connect(SIGNAL('triggered()'), self.downloadExportFile)
         menu.addAction('About').connect(SIGNAL('triggered()'), self.about)
@@ -301,10 +448,9 @@ class LTMainWindow(QWidget): #{{{1
     def setupScreen(self): #{{{2
         # Create the required objects
         self._tab      = QTableView()
-        self._model    = QSqlQueryModel()
         self._network  = QNetworkAccessManager(self)
         self._progress = QProgressDialog(self)
-        self._tab.setModel(self._model)
+        self._tab.setModel(QSqlQueryModel(self))
 
         try:
             self._database = LTDatabase()
@@ -338,12 +484,13 @@ class LTMainWindow(QWidget): #{{{1
 
     def loadData(self): #{{{2
         # Set the SQL query to use
-        col_widths = self._database.getQuery(self._model)
+        self._database.getQuery(self._tab.model())
 
         # Hide the ID column
         self._tab.setColumnHidden(0, True)
 
         # Set the column widths
+        col_widths = self._database.getColumnWidths()
         view_size = self._tab.viewport().width()
 
         var_size_count = len(col_widths)
@@ -383,7 +530,7 @@ class LTMainWindow(QWidget): #{{{1
         if filename != None:
             self.setupProgressBar('LoadingData')
             self._database.loadFile(filename, self._progress)
-            self._model.clear()
+            self.__tab.model().clear()
             self.loadData()
 
     def loginLT(self): #{{{2
@@ -444,15 +591,15 @@ class LTMainWindow(QWidget): #{{{1
         aboutWin.exec_()
 
     def changeColumnOrder(self, id, oldPos, newPos): #{{{2
-        col = self._model.record().fieldName(id)
+        col = self._tab.model().record().fieldName(id)
         self._database.changeColumnOrder(col, oldPos, newPos)
 
     def changeColumnSize(self, id, oldSize, newSize): #{{{2
-        col = self._model.record().fieldName(id)
+        col = self._tab.model().record().fieldName(id)
         self._database.changeColumnSize(col, newSize)
 
     def viewBookInfo(self, item): #{{{2
-        bookID = self._model.record(item.row()).value(0).toInt()[0]
+        bookID = self._tab.model().record(item.row()).value(0).toInt()[0]
 
         LTBookWindow(self, self._database, bookID).exec_()
 
@@ -478,7 +625,7 @@ class LTMainWindow(QWidget): #{{{1
 
                 self.setupProgressBar('Loading data')
                 self._database.loadStream(stream, self._progress)
-                self._model.clear()
+                self._tab.model().clear()
                 self.loadData()
             else:
                 self._loginBox.done(0)
@@ -495,3 +642,7 @@ class LTMainWindow(QWidget): #{{{1
             reply.deleteLater()
             self.emit(SIGNAL('downloadComplete()'))
 
+    def displayFilters(self): #{{{2
+        filterWin = LTFilterDefine(self, self._database)
+        if filterWin.exec_():
+            self.loadData()
