@@ -1,6 +1,7 @@
 #!/usr/bin/env python
 
 import sys
+import traceback
 
 from PySide.QtGui import QApplication, QMessageBox
 from LTgui import LTMainWindow
@@ -17,6 +18,6 @@ if __name__ == '__main__':
 
         app.exec_()
     except Exception, e:
-        QMessageBox.critical(None, 'Error', str(e))
+        QMessageBox.critical(None, 'Error', traceback.format_exc())
         
     sys.exit()
